package com.msignia.samples.threeds.backend.controllers;

import com.msignia.samples.threeds.backend.components.BrowserIPAddress;
import com.msignia.samples.threeds.backend.logger.TransactionLogger;
import com.msignia.samples.threeds.backend.models.AuthenticationMPIRequest;
import com.msignia.samples.threeds.backend.models.BrowserAuthenticationRequest;
import com.msignia.samples.threeds.backend.models.DeviceChannel;
import com.msignia.samples.threeds.backend.models.MobileAuthenticationRequest;
import com.msignia.samples.threeds.backend.services.AuthenticationServiceImpl;
import com.msignia.samples.threeds.backend.services.AuthenticationServiceMPI;
import com.msignia.threeds.mpi.authentication.AuthenticationResponse;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import javax.servlet.http.HttpServletRequest;
import java.nio.charset.StandardCharsets;
import java.util.Base64;


@RestController
@RequestMapping("/api")
public class AuthenticationController {

    private final Logger logger = LoggerFactory.getLogger(AuthenticationController.class);
    private final AuthenticationServiceImpl authenticationServiceImpl;
    private final AuthenticationServiceMPI authenticationServiceByMPI;
    private final TransactionLogger transactionLogger;

    public AuthenticationController(AuthenticationServiceImpl authenticationServiceImpl,
                                    AuthenticationServiceMPI authenticationServiceByMPI,
                                    TransactionLogger transactionLogger) {
        this.authenticationServiceImpl = authenticationServiceImpl;
        this.authenticationServiceByMPI = authenticationServiceByMPI;
        this.transactionLogger = transactionLogger;
    }

    @CrossOrigin(origins = "*")
    @PostMapping(path = "/v2/auth", consumes = "application/json", produces = "application/json")
    public ResponseEntity auth(HttpServletRequest httpServletRequest,
                               @RequestBody AuthenticationMPIRequest request) throws Exception {
        logger.debug("Auth method entered, request: {}", request);

        if (DeviceChannel.BRW.messageCode().equals(request.getAuthenticationRequestParameters().getDeviceChannel())) {
            request.getAuthenticationRequestParameters().setBrowserIP(new BrowserIPAddress(httpServletRequest).getIP());
            request.getAuthenticationRequestParameters().setBrowserAcceptHeader(
                    String.join("/",
                            httpServletRequest.getHeader("Accept"),
                            httpServletRequest.getHeader("Accept-Encoding"),
                            httpServletRequest.getHeader("Accept-Language")));
        }

        AuthenticationResponse result = authenticationServiceByMPI.perform(request);

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .headers(null)
                .payload(new JSONObject(request).toString())
                .build();
        transactionLogger.log(result.getThreeDSServerTransID(), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .headers(null)
                .payload(new JSONObject(result).toString())
                .build();

        transactionLogger.log(result.getThreeDSServerTransID(), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return ResponseEntity.ok().body(new JSONObject(result).toMap());
    }

    @PostMapping(path = "/v1/auth", consumes = "application/json", produces = "application/json")
    public String auth(@RequestBody MobileAuthenticationRequest mobileRequest) throws Exception {
        logger.debug("Auth method entered, mobileRequest: {}", mobileRequest);

        JSONObject result = authenticationServiceImpl.perform(mobileRequest);

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .headers(null)
                .payload(new JSONObject(mobileRequest).toString())
                .build();
        transactionLogger.log(result.getString("threeDSServerTransID"), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .headers(null)
                .payload(result.toString())
                .build();

        transactionLogger.log(result.getString("threeDSServerTransID"), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return result.toString();
    }

    @CrossOrigin(origins = "*")
    @PostMapping(path = "/v1/browser-auth", consumes = "application/json", produces = "application/json")
    public String auth(
            HttpServletRequest httpServletRequest,
            @RequestBody BrowserAuthenticationRequest browserRequest) throws Exception {

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(new JSONObject(browserRequest).toString())
                .headers(null)
                .build();
        transactionLogger.log(browserRequest.getThreeDSServerTransID(), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        browserRequest.setBrowserIP(new BrowserIPAddress(httpServletRequest).getIP());
        browserRequest.setBrowserAcceptHeader(
                String.join("/", String.join("/",
                        httpServletRequest.getHeader("Accept"),
                        httpServletRequest.getHeader("Accept-Encoding"),
                        httpServletRequest.getHeader("Accept-Language"))));

        JSONObject result = authenticationServiceImpl.perform(browserRequest);

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(result.toString())
                .headers(null)
                .build();

        transactionLogger.log(result.getString("threeDSServerTransID"), TransactionLogger.Flow.AREQ_ARES_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return result.toString();
    }

    @PostMapping(path = "/v1/3ds-method-notification", consumes = MediaType.APPLICATION_FORM_URLENCODED_VALUE, produces = MediaType.TEXT_HTML_VALUE)
    public String threeDsMethodNotification(String threeDSMethodData) {
        logger.debug("Browser notification method entered...");

        final byte[] threeDsMethodDataDecoded = Base64.getDecoder().decode(threeDSMethodData);
        final JSONObject data = new JSONObject(new String(threeDsMethodDataDecoded, StandardCharsets.UTF_8));
        logger.debug("Browser notification data: {}", data);

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(threeDSMethodData)
                .headers(null)
                .build();
        transactionLogger.log(data.getString("threeDSServerTransID"), TransactionLogger.Flow.THREE_DS_METHOD_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        String response = "<script>\n" +
                "  document.addEventListener(\"DOMContentLoaded\", function () {\n" +
                "    window.parent.postMessage({" +
                "\"type\":\"threeds-method-notification\"}," +
                // here should rather be a concrete targetOrigin,
                // see https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                "\"" + "*" + "\");\n" +
                "  });\n" +
                "</script>";

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(response)
                .headers(null)
                .build();
        transactionLogger.log(data.getString("threeDSServerTransID"), TransactionLogger.Flow.THREE_DS_METHOD_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return response;
    }

    @PostMapping(path = "/v1/browser-challenge-completion-notification", consumes = MediaType.APPLICATION_FORM_URLENCODED_VALUE, produces = MediaType.TEXT_HTML_VALUE)
    public String browserChallengeCompletionNotification(@RequestParam(name = "cres") String encodedCres) {
        logger.debug("Browser challenge completion notification method entered...");

        JSONObject cres = new JSONObject(new String(Base64.getDecoder().decode(encodedCres)));
        logger.debug("Decoded CRes: {}", cres.toString());

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(encodedCres)
                .headers(null)
                .build();
        transactionLogger.log(cres.getString("threeDSServerTransID"), TransactionLogger.Flow.CREQ_CRES_BROWSER_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        final JSONObject postMessage = new JSONObject();
        postMessage.put("type", "mSIGNIA-3DSv2-TransactionMessage");

        if (cres.has("transStatus")) {
            // CRes.transStatus can only be Y or N
            String transStatus = cres.getString("transStatus");
            logger.debug("TransStatus found: {}", transStatus);

            if ("Y".equals(transStatus)) {
                postMessage.put("outcome", "succeeded");
            } else {
                if (cres.has("challengeCancel") && "01".equals(cres.get("challengeCancel"))) {
                    postMessage.put("outcome", "cancelled");
                } else {
                    postMessage.put("outcome", "failed");
                }
            }
        } else {
            logger.debug("Erro Message Received");
            postMessage.put("outcome", "erred");
        }

        String response = getChallengeCompletionResponse(postMessage);

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(response)
                .headers(null)
                .build();
        transactionLogger.log(cres.getString("threeDSServerTransID"), TransactionLogger.Flow.CREQ_CRES_BROWSER_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return response;
    }

    @PostMapping(path = "/v2/browser-challenge-completion-notification", consumes = MediaType.APPLICATION_FORM_URLENCODED_VALUE, produces = MediaType.TEXT_HTML_VALUE)
    public String browserChallengeCompletionNotificationV2(@RequestParam(name = "cres") String encodedCres) {
        logger.debug("Browser challenge completion notification method entered...");

        JSONObject cres = new JSONObject(new String(Base64.getDecoder().decode(encodedCres)));
        logger.debug("Decoded CRes: {}", cres.toString());

        TransactionLogger.LogEntry entryRequest = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(encodedCres)
                .headers(null)
                .build();
        transactionLogger.log(cres.getString("threeDSServerTransID"), TransactionLogger.Flow.CREQ_CRES_BROWSER_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .to(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .build().getRoute(),
                entryRequest);

        final JSONObject postMessage = new JSONObject();
        postMessage.put("type", "threeds-challenge-notification");

        if (cres.has("transStatus")) {
            // CRes.transStatus can only be Y or N
            String transStatus = cres.getString("transStatus");
            logger.debug("TransStatus found: {}", transStatus);

            if ("Y".equals(transStatus)) {
                postMessage.put("outcome", "succeeded");
            } else {
                if (cres.has("challengeCancel") && "01".equals(cres.get("challengeCancel"))) {
                    postMessage.put("outcome", "cancelled");
                } else {
                    postMessage.put("outcome", "failed");
                }
            }
        } else {
            logger.debug("Erro Message Received");
            postMessage.put("outcome", "erred");
        }

        String response = getChallengeCompletionResponse(postMessage);

        TransactionLogger.LogEntry entryResponse = new TransactionLogger.LogEntry.Builder()
                .url(null)
                .payload(response)
                .headers(null)
                .build();
        transactionLogger.log(cres.getString("threeDSServerTransID"), TransactionLogger.Flow.CREQ_CRES_BROWSER_V1,
                new TransactionLogger.LogPath.Builder()
                        .from(TransactionLogger.Requestor.MERCHANT_BACKEND)
                        .to(TransactionLogger.Requestor.MERCHANT_CLIENT)
                        .build().getRoute(),
                entryResponse);

        return response;
    }

    private String getChallengeCompletionResponse(JSONObject postMessage) {
        return "<script>\n" +
                "  document.addEventListener(\"DOMContentLoaded\", function () {\n" +
                "    window.parent.postMessage(" + postMessage.toString() +
                // here should rather be a concrete targetOrigin,
                // see https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                ", \"" + "*" + "\");\n" +
                "  });\n" +
                "</script>";
    }
}
