package com.msignia.samples.threeds.backend.logger;

import okhttp3.*;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import java.util.Map;

@Component
public class TransactionLogger {

    private final Logger logger = LoggerFactory.getLogger(TransactionLogger.class);

    private final OkHttpClient okHttpClient;
    private final String transactionLoggingUrl;

    public TransactionLogger(OkHttpClient okHttpClient,
                             @Value("${app.transaction-logging-url}") String transactionLoggingUrl) {
        this.okHttpClient = okHttpClient;
        this.transactionLoggingUrl = transactionLoggingUrl;
    }

    public void log(String transactionId, Flow flow, String index, LogEntry logEntry) {
        JSONObject payload = new JSONObject();
        payload.put("flowId", flow.name());
        payload.put("index", index);
        payload.put("data", new JSONObject(logEntry));

        logger.debug("Sending log: {}", payload.toString(4));

        try {
            JSONObject response = post(transactionId, payload, transactionLoggingUrl);
            logger.debug("Received response: {}", response.toString(4));
        } catch(Exception e) {
            logger.warn("Failed to send request, error: ", e);
        }
    }

    public enum Flow {
        SUPPORTED_VERSIONS_V1,
        THREE_DS_METHOD_V1,
        AREQ_ARES_V1,
        CREQ_CRES_APP_V1,
        CREQ_CRES_BROWSER_V1,
        RREQ_RRES_V1,
        PREQ_PRES_V1 //???
    }

    public enum Requestor {
        MERCHANT_CLIENT("MC"),
        MERCHANT_BACKEND("MB"),
        THREE_DS_SERVER("3DSS"),
        DS("DS"),
        ACS("ACS"),
        BEHAVIO_SENSE("BSE");

        public final String requestor;

        Requestor(String requestor) {
            this.requestor = requestor;
        }
    }

    public static class LogPath {
        String route;

        public static class Builder {
            public Requestor from;
            public Requestor to;
            public String index;

            public LogPath build() {
                LogPath logPath = new LogPath();

                StringBuilder stringBuilder = new StringBuilder();

                stringBuilder.append(from.requestor).append(":").append(to.requestor);

                if (!StringUtils.isEmpty(index)) {
                    stringBuilder.append("[").append(index).append("]");
                }

                logPath.route = stringBuilder.toString();

                return logPath;
            }

            public Builder from(Requestor requestor) {
                this.from = requestor;
                return this;
            }

            public Builder to(Requestor requestor) {
                this.to = requestor;
                return this;
            }

            public Builder index(String value) {
                this.index = value;
                return this;
            }
        }

        public String getRoute() {
            return route;
        }

        public void setRoute(String route) {
            this.route = route;
        }
    }

    public static class LogEntry {
        private String url;
        private String method;
        private String payload;
        private Map<String, String> headers;

        private LogEntry() {}

        public static class Builder {
            private String url;
            private String method = "POST";
            private String payload;
            private Map<String, String> headers;

            public LogEntry build() {
                LogEntry result = new LogEntry();

                result.url = this.url;
                result.method = this.method;
                result.payload = this.payload;
                result.headers = this.headers;

                return result;
            }

            public Builder url(String url) {
                this.url = url;
                return this;
            }

            public Builder method(String method) {
                this.method = method;
                return this;
            }

            public Builder payload(String payload) {
                this.payload = payload;
                return this;
            }

            public Builder headers(Map<String, String> headers) {
                this.headers = headers;
                return this;
            }
        }

        public String getUrl() {
            return url;
        }

        public void setUrl(String url) {
            this.url = url;
        }

        public String getMethod() {
            return method;
        }

        public void setMethod(String method) {
            this.method = method;
        }

        public String getPayload() {
            return payload;
        }

        public void setPayload(String payload) {
            this.payload = payload;
        }

        public Map<String, String> getHeaders() {
            return headers;
        }

        public void setHeaders(Map<String, String> headers) {
            this.headers = headers;
        }
    }

    private JSONObject post(String transactionId, JSONObject payload, String url) throws Exception {
        RequestBody body = RequestBody.create(payload.toString(), MediaType.parse("application/json"));
        logger.debug("Sending log: {}/{}", url, payload.toString(4));

        Request.Builder requestBuilder = new Request.Builder().url(url + "/" + transactionId).post(body);

        Request request = requestBuilder.build();

        try (Response response = okHttpClient.newCall(request).execute()) {
            String result = response.body().string();
            return StringUtils.isEmpty(result) ? new JSONObject() : new JSONObject(result);
        }
    }
}
