package com.msignia.samples.threeds.backend.services;

import com.msignia.samples.threeds.backend.models.AuthenticationRequestParameters;
import com.msignia.samples.threeds.backend.models.BrowserAuthenticationRequest;
import com.msignia.samples.threeds.backend.models.MobileAuthenticationRequest;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

@Service
public class AuthenticationServiceImpl implements AuthenticationService {

    private final Logger logger = LoggerFactory.getLogger(AuthenticationServiceImpl.class);

    private final ThreeDsServerClient threeDsServerClient;
    private final String threeDSRequestorID;

    public AuthenticationServiceImpl(ThreeDsServerClient threeDsServerClient,
                                     @Value("${app.three-ds-requestor-id}") String threeDSRequestorID) {
        this.threeDsServerClient = threeDsServerClient;
        this.threeDSRequestorID = threeDSRequestorID;
    }

    public JSONObject perform(MobileAuthenticationRequest mobileRequest) throws Exception {
        logger.debug("perform#MobileAuthenticationRequest entered...");
        JSONObject requestBody = prepareRequest(mobileRequest);
        return threeDsServerClient.authenticate(requestBody);
    }

    public JSONObject perform(BrowserAuthenticationRequest browserRequest) throws Exception {
        logger.debug("perform#BrowserAuthenticationRequest entered...");
        JSONObject requestBody = prepareRequest(browserRequest);
        return threeDsServerClient.authenticate(requestBody);
    }

    private JSONObject prepareRequest(BrowserAuthenticationRequest browserRequest) {
        JSONObject result = new JSONObject();

        result.put("acctNumber", browserRequest.getAcctNumber());
        result.put("browserAcceptHeader", browserRequest.getBrowserAcceptHeader());
        result.put("browserIP", browserRequest.getBrowserIP());
        result.put("browserJavaEnabled", browserRequest.isBrowserJavaEnabled());
        result.put("browserJavascriptEnabled", browserRequest.isBrowserJavascriptEnabled());
        result.put("browserLanguage", browserRequest.getBrowserLanguage());
        result.put("browserColorDepth", browserRequest.getBrowserColorDepth());
        result.put("browserScreenHeight", browserRequest.getBrowserScreenHeight());
        result.put("browserScreenWidth", browserRequest.getBrowserScreenWidth());
        result.put("browserTZ", browserRequest.getBrowserTZ());
        result.put("browserUserAgent", browserRequest.getBrowserUserAgent());

        result.put("threeDSCompInd", browserRequest.getThreeDSCompInd());
        result.put("notificationURL", browserRequest.getNotificationURL());
        result.put("threeDSServerTransID", browserRequest.getThreeDSServerTransID());
        result.put("challengeWindowSize", "03"); // 03 = 500 x 600

        result.put("deviceChannel", "02"); // 02 = Browser (BRW) versus App-based

        result.put("messageVersion", "2.2.0");
        result.put("dsIdentifier", browserRequest.getDsIdentifier());

        addCommonParametersToRequest(result);

        return result;
    }

    private JSONObject prepareRequest(MobileAuthenticationRequest mobileRequest) {
        AuthenticationRequestParameters authParams = mobileRequest.getAuthRequestParameters();

        JSONObject result = new JSONObject();
        result.put("acctNumber", mobileRequest.getAcctNumber());

        // copying `authParams` into the `result`, note not all the params can be copied literally:
        result.put("messageVersion", authParams.getMessageVersion());
        result.put("sdkEncData", authParams.getDeviceData());
        result.put("sdkTransID", authParams.getSdkTransactionID());
        result.put("sdkAppID", authParams.getSdkAppID());
        result.put("sdkReferenceNumber", authParams.getSdkReferenceNumber());
        result.put("sdkEphemPubKey", new JSONObject(authParams.getSdkEphemeralPublicKey()));

        // these two should normally be set by the mobile app (not SDK!)
        result.put("deviceRenderOptions", new JSONObject(mobileRequest.getDeviceRenderOptions()));
        result.put("sdkMaxTimeout", mobileRequest.getSdkMaxTimeout());

        result.put("deviceChannel", "01"); // 01 = App-based (APP) versus Browser

        result.put("dsIdentifier", mobileRequest.getDsIdentifier());

        addCommonParametersToRequest(result);

        return result;
    }

    private void addCommonParametersToRequest(JSONObject request) {

        request.put("threeDSRequestorAuthenticationInd", "01"); // 01 - Payment transaction
        request.put("threeDSRequestorID", threeDSRequestorID); // the backend may not know this, in this case 3DSS will set it
        request.put("threeDSRequestorName", "mSIGNIA Test Merchant");
        request.put("threeDSRequestorURL", "https://merchant.msignia.com");

        request.put("acctType", "01");
        request.put("acquirerBIN", "444444");
        request.put("acquirerMerchantID", "550e8400");

        request.put("merchantCountryCode", "840");
        request.put("merchantName", "mSIGNIA Test Merchant");
        request.put("messageCategory", "01"); // 01 = Payment Authentication (PA)
        request.put("transType", "01"); // 01 = Goods/ Service Purchase
        request.put("mcc", "5411");

        // Cardholder and purchase information
        request.put("addrMatch", "N");
        request.put("cardExpiryDate", "2001");

        request.put("billAddrCity", "Chicago");
        request.put("billAddrCountry", "004");
        request.put("billAddrLine1", "Main Street");
        request.put("billAddrLine2", "15");
        request.put("billAddrPostCode", "88888");
        request.put("billAddrState", "IL");

        request.put("email", "sample.app@msignia.com");
        request.put("homePhone", getPhone("7777777777"));
        request.put("mobilePhone", getPhone("8888888888"));
        request.put("cardholderName", "Pavlo V. Lysov");

        request.put("shipAddrCity", "Nashville");
        request.put("shipAddrCountry", "004");
        request.put("shipAddrLine1", "Belmont blvd");
        request.put("shipAddrLine2", "2994");
        request.put("shipAddrPostCode", "77777");
        request.put("shipAddrState", "TN");

        request.put("purchaseAmount", "4595");
        request.put("purchaseCurrency", "840");
        request.put("purchaseExponent", "2");
        request.put("purchaseDate", "20200327103740");
    }

    private JSONObject getPhone(String subscriber) {
        JSONObject result = new JSONObject();
        result.put("cc", "1");
        result.put("subscriber", subscriber);

        return result;
    }
}
