package com.msignia.samples.threeds.backend.services;

import com.msignia.samples.threeds.backend.models.AuthenticationMPIRequest;
import com.msignia.samples.threeds.backend.models.AuthenticationRequestMPIParameters;
import com.msignia.samples.threeds.backend.models.DeviceChannel;
import com.msignia.threeds.mpi.authentication.AuthenticationParameters;
import com.msignia.threeds.mpi.authentication.AuthenticationResponse;
import com.msignia.threeds.mpi.authentication.models.PhoneNumber;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

@Service
public class AuthenticationServiceMPI implements AuthenticationService {

    private final Logger logger = LoggerFactory.getLogger(AuthenticationServiceMPI.class);

    private final ThreeDsServerClientByMPI threeDsServerClient;
    private final String threeDSRequestorID;

    public AuthenticationServiceMPI(ThreeDsServerClientByMPI threeDsServerClient,
                                    @Value("${app.three-ds-requestor-id}") String threeDSRequestorID) {
        this.threeDsServerClient = threeDsServerClient;
        this.threeDSRequestorID = threeDSRequestorID;
    }

    public AuthenticationResponse perform(AuthenticationMPIRequest mobileRequest) throws Exception {
        logger.debug("perform#MobileAuthenticationRequest entered...");
        AuthenticationParameters authParams =
                DeviceChannel.APP.messageCode().equals(mobileRequest.getAuthenticationRequestParameters().getDeviceChannel()) ?
                prepareAppParams(mobileRequest) :
                prepareBrowserParams(mobileRequest);

        return threeDsServerClient.authenticate(authParams);
    }

    private AuthenticationParameters prepareBrowserParams(AuthenticationMPIRequest request) {
        AuthenticationParameters params = new AuthenticationParameters();

        String cardId = request.getAuthenticationRequestParameters().getCardID();
        params.setAcctNumber(StringUtils.isEmpty(cardId) ? request.getPurchaseInfo().getCardID() : cardId);

        params.setBrowserAcceptHeader(request.getAuthenticationRequestParameters().getBrowserAcceptHeader());
        params.setBrowserIP(request.getAuthenticationRequestParameters().getBrowserIP());
        params.setBrowserJavaEnabled(request.getAuthenticationRequestParameters().getBrowserJavaEnabled());
        params.setBrowserJavascriptEnabled(request.getAuthenticationRequestParameters().getBrowserJavascriptEnabled());
        params.setBrowserLanguage(request.getAuthenticationRequestParameters().getBrowserLanguage());
        params.setBrowserColorDepth(request.getAuthenticationRequestParameters().getBrowserColorDepth());
        params.setBrowserScreenHeight(request.getAuthenticationRequestParameters().getBrowserScreenHeight());
        params.setBrowserScreenWidth(request.getAuthenticationRequestParameters().getBrowserScreenWidth());
        params.setBrowserTZ(request.getAuthenticationRequestParameters().getBrowserTZ());
        params.setBrowserUserAgent(request.getAuthenticationRequestParameters().getBrowserUserAgent());

        params.setThreeDSCompInd(request.getAuthenticationRequestParameters().getThreeDSCompInd());
        params.setNotificationURL(request.getAuthenticationRequestParameters().getNotificationURL());
        params.setThreeDSServerTransID(request.getAuthenticationRequestParameters().getThreeDSServerTransID());
        params.setChallengeWindowSize(request.getAuthenticationRequestParameters().getChallengeWindowSize()); // 03 = 500 x 600

        params.setDeviceChannel(request.getAuthenticationRequestParameters().getDeviceChannel()); // 02 = Browser (BRW) versus App-based
        params.setMessageVersion(request.getAuthenticationRequestParameters().getMessageVersion());

        addCommonParametersToRequest(params, request);

        return params;
    }

    private AuthenticationParameters prepareAppParams(AuthenticationMPIRequest mobileRequest) {
        AuthenticationRequestMPIParameters authRequestParameters = mobileRequest.getAuthenticationRequestParameters();
        String cardId = mobileRequest.getAuthenticationRequestParameters().getCardID();

        AuthenticationParameters authenticationParameters = new AuthenticationParameters();
        authenticationParameters.setAcctNumber(StringUtils.isEmpty(cardId) ? mobileRequest.getPurchaseInfo().getCardID() : cardId);
        authenticationParameters.setMessageVersion(authRequestParameters.getMessageVersion());
        authenticationParameters.setSdkEncData(authRequestParameters.getDeviceData());
        authenticationParameters.setSdkTransID(authRequestParameters.getSdkTransactionID());
        authenticationParameters.setSdkAppID(authRequestParameters.getSdkAppID());
        authenticationParameters.setSdkReferenceNumber(authRequestParameters.getSdkReferenceNumber());
        authenticationParameters.setSdkEphemPubKey(authRequestParameters.getSdkEphemeralPublicKey());

        // these two should normally be set by the mobile app (not SDK!)
        authenticationParameters.setDeviceRenderOptions(mobileRequest.getAuthenticationRequestParameters().getDeviceRenderOptions());
        authenticationParameters.setSdkMaxTimeout(mobileRequest.getAuthenticationRequestParameters().getSdkMaxTimeout());
        authenticationParameters.setDeviceChannel(mobileRequest.getAuthenticationRequestParameters().getDeviceChannel()); // 01 = App-based (APP) versus Browser

        addCommonParametersToRequest(authenticationParameters, mobileRequest);

        return authenticationParameters;
    }

    private void addCommonParametersToRequest(AuthenticationParameters params, AuthenticationMPIRequest mobileRequest) {

        params.setThreeDSRequestorAuthenticationInd("01"); // 01 - Payment transaction
        params.setThreeDSRequestorID(threeDSRequestorID); // the backend may not know this, in this case 3DSS will set it
        params.setThreeDSRequestorName("mSIGNIA Test Merchant");
        params.setThreeDSRequestorURL("https://merchant.msignia.com");

        params.setAcctType("01");
        params.setAcquirerBIN("444444");
        params.setAcquirerMerchantID("550e8400");

        params.setMerchantCountryCode("840");
        params.setMerchantName("mSIGNIA Test Merchant");
        params.setMessageCategory(mobileRequest.getAuthenticationRequestParameters().getMessageCategory()); // 01 = Payment Authentication (PA)
        params.setTransType("01"); // 01 = Goods/ Service Purchase
        params.setMcc("5411");

        // Cardholder and purchase information
        params.setAddrMatch("N");
        params.setAddrMatch("N");
        params.setCardExpiryDate("2001");

        params.setBillAddrCity("Chicago");
        params.setBillAddrCountry("004");
        params.setBillAddrLine1("Main Street");
        params.setBillAddrLine2("15");
        params.setBillAddrPostCode("88888");
        params.setBillAddrState("IL");

        params.setEmail(mobileRequest.getPurchaseInfo().getEmail());
        params.setHomePhone(getPhone("7777777777"));
        params.setMobilePhone(getPhone("8888888888"));
        params.setCardholderName("Pavlo V. Lysov");

        params.setShipAddrCity("Nashville");
        params.setShipAddrCountry("004");
        params.setShipAddrLine1("Belmont blvd");
        params.setShipAddrLine2("2994");
        params.setShipAddrPostCode("77777");
        params.setShipAddrState("TN");

        params.setPurchaseAmount(mobileRequest.getPurchaseInfo().getPurchaseAmount());
        params.setPurchaseCurrency(mobileRequest.getPurchaseInfo().getPurchaseCurrency());
        params.setPurchaseExponent(mobileRequest.getPurchaseInfo().getPurchaseExponent());
        params.setPurchaseDate(mobileRequest.getPurchaseInfo().getPurchaseDate());

        params.setDsIdentifier(mobileRequest.getAuthenticationRequestParameters().getDsIdentifier());
    }

    private PhoneNumber getPhone(String subscriber) {
        return new PhoneNumber("1", subscriber);
    }
}
